from enum import IntEnum, unique

from sqlalchemy import and_, func
from sqlalchemy.orm import Query, Session, aliased

import model as m
from cas.data import Session as SessionFunc
from cas.utils import has_value


@unique
class BimDataModelType(IntEnum):
    Architectural = 1 << 0
    Structural = 1 << 1
    Mep = 1 << 2
    Custom = 1 << 3


@unique
class BimDataModelPhase(IntEnum):
    PreConstruction = 0
    Construction = 1
    PostConstruction = 2
    Refurbishment = 3


@unique
class BimDataModelState(IntEnum):
    Initialising = 0,
    Ready = 1,
    Updating = 2,
    Error = 3


@unique
class BimDataModelFileType(IntEnum):
    Unknown = 0
    ThreeDModel = 1
    TwoDModel = 2
    Ifc = 3
    Csv = 4
    BlueprintJson = 5


class BimDataModelRepository:

    def __init__(self, sql_session: Session):
        self.__session = sql_session

    def get_model_file_name(self, project_id: str, organisation_id: str,
                            file_type: BimDataModelFileType = BimDataModelFileType.Ifc):
        model_query = self.__default_model_query(project_id, organisation_id)
        file = self.__generate_model_file_query(model_query, filter_file_type=file_type).first()
        if file is None:
            return None

        file_name = f'{file.ModelId}_{file.VersionNumber}'
        if has_value(file.FileNameSuffix):
            file_name = f'{file_name}_{file.FileNameSuffix}'

        return file_name

    def find_default_model(self, project_id: str, organisation_id: str) -> m.BimDataModel:
        return self.__default_model_query(project_id, organisation_id).first()

    def __default_model_query(self, project_id: str, organisation_id: str) -> Query:
        return self.__generate_model_query(fileter_project_id=project_id,
                                           filter_organisation_id=organisation_id,
                                           filter_model_type=BimDataModelType.Architectural,
                                           filter_model_state=BimDataModelState.Ready,
                                           filter_by_default_model=True)

    def __generate_model_query(self,
                               fileter_project_id: str,
                               filter_organisation_id: str,
                               filter_model_id: str = None,
                               filter_model_type: BimDataModelType = None,
                               filter_model_phase: BimDataModelPhase = None,
                               filter_model_state: BimDataModelState = None,
                               filter_by_default_model: bool = False) -> Query:
        """Will apply the query parameters to the Query. If the query is None, a new query will be generated from the
        Session."""
        result = self.__session.query(m.BimDataModel)

        if has_value(fileter_project_id):
            result = result.filter(m.BimDataModel.ProjectId == fileter_project_id)

        if has_value(filter_organisation_id):
            result = result.filter(m.BimDataModel.OrganisationId == filter_organisation_id)

        if filter_model_type is not None:
            result = result.filter(m.BimDataModel.ModelType == int(filter_model_type))

        if filter_model_phase is not None:
            result = result.filter(m.BimDataModel.ModelPhase == int(filter_model_phase))

        if filter_model_state is not None:
            result = result.filter(m.BimDataModel.ModelState == int(filter_model_state))

        if has_value(filter_model_id):
            result = result.filter(m.BimDataModel.ModelId == filter_model_id)
        elif filter_by_default_model:
            result = result.filter(m.BimDataModel.IsDefault)

        return result

    def __generate_model_version_query(self,
                                       model_query: Query,
                                       filter_version_number: int = None) -> Query:
        sub_q = model_query.subquery(name='bim_model')
        result = self.__session.query(m.BimDataModelVersion).join(sub_q,
                                                                  m.BimDataModelVersion.ModelId == sub_q.c.ModelId)

        if filter_version_number is not None:
            result = result.filter(m.BimDataModelVersion.VersionNumber == filter_version_number)

        return result

    def __generate_model_file_query(self,
                                    model_query: Query,
                                    filter_file_type: BimDataModelFileType = None,
                                    filter_latest_file: bool = True) -> Query:
        model_subquery = model_query.subquery(name='bim_model')
        parent_alias = aliased(m.BimDataModelVersionFile, name='bim_file')

        result = self.__session.query(parent_alias) \
            .join(model_subquery, parent_alias.ModelId == model_subquery.c.ModelId)

        if filter_file_type is not None:
            result = result.filter(parent_alias.BimModelFileType == int(filter_file_type))

        if filter_latest_file:
            # If we only need the latest file, we'll filter it my the max version
            child_alias = aliased(m.BimDataModelVersionFile, name='version_file')
            max_query = self.__session.query(func.max(child_alias.VersionNumber)) \
                .filter(and_(child_alias.ModelId == parent_alias.ModelId,
                             child_alias.BimModelFileType == parent_alias.BimModelFileType))

            result = result.filter(parent_alias.VersionNumber == max_query)

        return result


if __name__ == '__main__':
    with SessionFunc() as session:
        repo = BimDataModelRepository(sql_session=session)
        model = repo.get_model_file_name('11220000-4800-acde-c6da-08da1619d625', 'CAS')

        print('Found', model)



import json
import logging
import os
import socket
import uuid
from json import JSONDecodeError
from multiprocessing import current_process
from threading import Thread, get_ident

import pika
import sys
from pika.exceptions import ChannelWrongStateError, ConnectionWrongStateError

from cas import utils

_LOGGER = logging.getLogger('cas.acabim.messaging')

EXPECTED_MSG_TYPE = {}  # fill in with the expected msg types that needs to be sent to MassTransit


class MessagingHandler:
    def __init__(self, connection_name='cas.messaging.PythonClient'):
        self.current_status = {}
        self.connection_name = connection_name
        self.__pending_listeners = {}
        self.__connected_callbacks = []
        self.__disconnected_callbacks = []
        self.__create_connection()
        self.__create_thread()

    def __del__(self):
        if self.is_connected():
            self.stop()

    def __create_connection(self):
        if not hasattr(self, '__connection'):
            user_name = os.getenv('ACABIM_MESSAGING_USERNAME', 'guest')
            password = os.getenv('ACABIM_MESSAGING_PASSWORD', 'guest')
            host = os.getenv('ACABIM_MESSAGING_HOST', 'localhost')
            credentials = pika.PlainCredentials(user_name, password)
            params = pika.ConnectionParameters(host=host, credentials=credentials, retry_delay=10,
                                               connection_attempts=10,
                                               client_properties={
                                                   'connection_name': self.connection_name,
                                                   'hostname': socket.gethostname(),
                                                   'python_version': sys.version,
                                                   'process_id': os.getpid()
                                               })
            self.__connection = pika.SelectConnection(params, on_open_callback=self.__on_connected,
                                                      on_close_callback=self.__on_closed)

    def __create_thread(self):
        if not hasattr(self, '_MessagingHandler__bk_thread'):
            self.__bk_thread = Thread(target=self.__connect, name='cas.messaging.ConnectionHandler')
            self.__bk_thread.daemon = False

    def is_connected(self):
        try:
            return self.__connection.is_open and self.__channel.is_open
        except AttributeError:
            return False

    def start(self):
        if not self.is_connected():
            self.__create_connection()
            self.__create_thread()
            self.__bk_thread.start()

    def stop(self):
        self.__remove_connection()

    def listen(self, queue_name, callback):
        if callback is None:
            raise TypeError('callback cannot be none')

        self.__pending_listeners[queue_name] = callback
        self.__channel.exchange_declare(queue_name, exchange_type='fanout', durable=True,
                                        callback=lambda ex: self.__on_exchange_declared(ex, queue_name))

    def wait_for_connection(self):
        _LOGGER.debug('Waiting for connection')
        while True:
            if self.is_connected():
                _LOGGER.debug('Connection established')
                break

    def respond(self, req_queue_name, res_message_body, req_message):
        """ Respond with a message for a received request """
        if not MessagingHandler.can_respond(req_message):
            _LOGGER.critical('Cannot respond to msg %s, No-Respond-Address-Found', json.dumps(req_message))
        else:
            res_queue = MessagingHandler.__get_response_queue_name(req_message)
            body = MessagingHandler.__create_response_message(req_queue_name, res_message_body, req_message)
            try:
                self.__channel.basic_publish(exchange=res_queue, routing_key='', body=body,
                                             properties=MessagingHandler.__get_header_properties())
                _LOGGER.debug(f'Message published on queue {res_queue}: Type: RespondSendMessage')
            except ChannelWrongStateError as channel_error:
                _LOGGER.exception(channel_error, 'Unable to send message - Channel not open')
            except AttributeError:
                _LOGGER.error('No connection found to respond with the message')

    def send_message(self, queue, res_message_body, req_message):
        try:
            body = MessagingHandler.__create_response_message(queue, res_message_body, req_message)
            try:
                self.__channel.basic_publish(exchange=queue, routing_key='', body=body,
                                             properties=MessagingHandler.__get_header_properties())
                _LOGGER.debug(f'Message published on queue {queue}: Type: BasicSendMessage')
            except ChannelWrongStateError as exc:
                _LOGGER.exception(exc, 'Unable to send message - Channel not open')
        except AttributeError as ex:
            _LOGGER.exception(ex)

    def send_fault_message(self, queue, exception, stack_trace, response, req_message):
        try:
            body = MessagingHandler.__create_fault_message(queue, str(exception), stack_trace, response, req_message)
            res_queue = MessagingHandler.__get_fault_queue_name(queue)
            try:
                self.__channel.basic_publish(exchange=res_queue, routing_key='', body=body,
                                             properties=MessagingHandler.__get_header_properties())
                _LOGGER.debug(f'Message published on queue {res_queue}: Type: FaultSendMessage')
            except ChannelWrongStateError as channel_error:
                _LOGGER.exception(channel_error, 'Unable to send message - Channel not open')
        except AttributeError as a_error:
            _LOGGER.exception(a_error, 'No connection found')

    def add_connected_callback(self, callback):
        self.__connected_callbacks.append(callback)

    def add_disconnected_callback(self, callback):
        self.__disconnected_callbacks.append(callback)

    def __connect(self):
        _LOGGER.info('Starting connection')
        try:
            self.__connection.ioloop.start()
        except KeyboardInterrupt:
            self.__connection.close()

    def __remove_connection(self):
        """ Removes the connection and the channel after closing it """
        try:
            _LOGGER.debug('Channel going down')
            self.__channel.close()
            del self.__channel
        except AttributeError:
            _LOGGER.warning('Unable to close channel - Channel not found')
        except ChannelWrongStateError:
            _LOGGER.warning('Unable to close channel - Channel may have already been closed')
            del self.__channel

        try:
            _LOGGER.debug('Connection going down')
            self.__connection.ioloop.add_callback_threadsafe(self.__connection.ioloop.stop)
            self.__connection.close()
            del self.__connection
        except AttributeError:
            _LOGGER.warning('Unable to remove connection - Connection not found')
        except ConnectionWrongStateError:
            _LOGGER.warning('Unable to close connection - connection may have already been closed')
            del self.__connection

        try:
            _LOGGER.info('Waiting for the background thread to exit')
            if get_ident() != self.__bk_thread.ident:
                self.__bk_thread.join()
            del self.__bk_thread
            _LOGGER.info('Background thread removed')
        except AttributeError:
            pass

    def __on_connected(self, opened_connection):
        _LOGGER.info('Connection Opened with Host: %s', opened_connection.params.host)
        self.__connection.channel(on_open_callback=self.__on_channel_open)

    def __on_closed(self, closed_connection, exception):
        _LOGGER.exception(exception)
        _LOGGER.critical('Connection closed with Host: %s', closed_connection.params.host)
        self.__remove_connection()
        for callback in self.__disconnected_callbacks:
            callback(self)
        self.__disconnected_callbacks.clear()

    def __on_channel_open(self, opened_channel):
        _LOGGER.info('Channel open with Host: %s, Number: %s', opened_channel.connection.params.host,
                     opened_channel.channel_number)
        self.__channel = opened_channel
        for call in self.__connected_callbacks:
            call(self)
        self.__connected_callbacks.clear()

    def __on_exchange_declared(self, exchange, exchange_name):
        _LOGGER.debug('Exchange Declared - Message: %s, ExchangeName: %s', exchange.method.NAME, exchange_name)
        if exchange.method.NAME == 'Exchange.DeclareOk':
            _LOGGER.debug('Declaring Queue: %s', exchange_name)
            self.__channel.queue_declare(queue=exchange_name, durable=True,
                                         callback=lambda result: self.__on_queue_declared(result, exchange_name))
        else:
            _LOGGER.critical('Invalid exchange message - Unable to declare queue %s', exchange_name)
            raise Exception('Invalid Response when declaring exchange \'{0}\', Received: {1} '
                            .format(exchange_name, exchange.method.NAME))

    def __on_queue_declared(self, queue, queue_name):
        _LOGGER.debug('Queue Declared - Message: %s, QueueName: %s', queue.method.NAME, queue_name)
        if queue.method.NAME == 'Queue.DeclareOk':
            _LOGGER.debug('Binding Queue %s to exchange %s', queue_name, queue_name)
            self.__channel.queue_bind(exchange=queue_name, queue=queue_name, routing_key='',
                                      callback=lambda result: self.__on_queue_ready(result, queue_name))
        else:
            _LOGGER.critical('Invalid exchange message - Unable to bind queue %s', queue_name)
            raise Exception('Invalid Response when binding queue \'{0}\', Received: {1}'
                            .format(queue_name, queue.method.NAME))

    def __on_queue_ready(self, queue, queue_name):
        if queue.method.NAME == 'Queue.BindOk':
            _LOGGER.info('Queue Ready: %s', queue_name)
            if queue_name in self.__pending_listeners:  # if we have a pending listener that means we have to listen
                _LOGGER.info('Listener found for Queue: %s', queue_name)
                self.__channel.basic_consume(queue=queue_name, auto_ack=True,
                                             on_message_callback=lambda channel, method, properties, body:
                                             self.__on_new_message_received(queue_name, body))

    def __on_new_message_received(self, queue_name, body):
        content = MessagingHandler.__verify_and_get_message_content(body)
        if queue_name in self.__pending_listeners:
            self.__pending_listeners[queue_name](content)

    @staticmethod
    def can_respond(message):
        return 'responseAddress' in message

    @staticmethod
    def __verify_and_get_message_content(body):
        try:
            return json.loads(body)
        except JSONDecodeError:
            _LOGGER.critical('Unhandled Body content: Invalid Message - Unable to parse JSON')
            _LOGGER.critical('Body: "%s"', body.decode('utf8'))
            _LOGGER.exception(JSONDecodeError)

    @staticmethod
    def __get_response_queue_name(req_message_content):
        return req_message_content['responseAddress'].split('/').pop().split('?')[0]

    @staticmethod
    def __get_fault_queue_name(expected_msg_type):
        return 'MassTransit:Fault--{0}--'.format(expected_msg_type)

    @staticmethod
    def __create_response_message(queue, response_message_content, req_message):
        res_message = {'messageId': str(uuid.uuid4())}
        if req_message is not None and 'requestId' in req_message:
            res_message['requestId'] = req_message['requestId']

        if req_message is not None and 'conversationId' in req_message:
            res_message['conversationId'] = req_message['conversationId']
        else:
            res_message['conversationId'] = str(uuid.uuid4())

        if req_message is not None and 'initiatorId' in req_message:
            res_message['initiatorId'] = req_message['initiatorId']
        else:
            res_message['initiatorId'] = str(uuid.uuid4())

        if queue in EXPECTED_MSG_TYPE:
            res_message['messageType'] = ['urn:message:{0}'.format(EXPECTED_MSG_TYPE[queue])]
        else:
            res_message['messageType'] = ['urn:message:{0}'.format(queue)]
            _LOGGER.warning('No message_type found, Using the queue name')

        res_message['message'] = response_message_content
        res_message['sentTime'] = utils.current_time_str()
        res_message['headers'] = {}
        res_message['host'] = MessagingHandler.__get_host_info()
        return bytes(json.dumps(res_message), 'utf8')

    @staticmethod
    def __create_fault_message(queue, exception_msg, stack_trace, response, req_message):
        res_message = {'messageId': str(uuid.uuid4())}
        if req_message is not None and 'requestId' in req_message:
            res_message['requestId'] = req_message['requestId']

        if req_message is not None and 'conversationId' in req_message:
            res_message['conversationId'] = req_message['conversationId']
        else:
            res_message['conversationId'] = str(uuid.uuid4())

        if req_message is not None and 'initiatorId' in req_message:
            res_message['initiatorId'] = req_message['initiatorId']
        else:
            res_message['initiatorId'] = str(uuid.uuid4())

        if queue in EXPECTED_MSG_TYPE:
            res_message['messageType'] = ['urn:message:MassTransit:Fault[[{0}]]'.format(EXPECTED_MSG_TYPE[queue]),
                                          'urn:message:MassTransit:Fault']
        else:
            res_message['messageType'] = ['urn:message:MassTransit:Fault[[{0}]]'.format(queue),
                                          'urn:message:MassTransit:Fault']
            _LOGGER.warning('No message_type found, Using the queue name')

        fault_message = {
            'faultId': str(uuid.uuid4()),
            'timestamp': utils.current_time_str(),
            'message': response,
            'exceptions': [
                {
                    'exceptionType': 'ComplianceAuditSystems.Acabim.Exceptions.PythonException',
                    'stackTrace': stack_trace,
                    'message': exception_msg
                }
            ],
            'host': MessagingHandler.__get_host_info()
        }

        if req_message is not None and 'messageId' in req_message:
            fault_message['faultedMessageId'] = req_message['messageId']

        if queue in EXPECTED_MSG_TYPE:
            fault_message['faultMessageTypes'] = ['urn:message:{0}'.format(EXPECTED_MSG_TYPE[queue])]

        res_message['message'] = fault_message
        res_message['sentTime'] = utils.current_time_str()
        res_message['headers'] = {}
        res_message['host'] = MessagingHandler.__get_host_info()
        return bytes(json.dumps(res_message), 'utf8')

    @staticmethod
    def __get_host_info():
        return {'machineName': socket.gethostname(),
                'processName': f'ifcparser ({current_process().name}-{Thread.name})',
                'processId': os.getpid(),
                'pythonVersion': sys.version
                }

    @staticmethod
    def __get_header_properties():
        return pika.BasicProperties(content_type='application/vnd.masstransit+json')


def encode_file(path: str, content=None, ext=None, mime_type=None):
    size = os.path.getsize(path) if path else len(content) if content else 0

    if mime_type is None:
        from mimetypes import guess_type
        mime_type = guess_type(path)[0]

    if ext is None and path and not path.isspace():
        exts = os.path.splitext(path)
        ext = exts[1].lstrip('.') if len(exts) == 2 else 'txt'  # assume file type as txt if no type is found

    from base64 import b64encode
    data = None
    if path and not path.isspace():
        with open(path, 'rb') as file:
            data = b64encode(file.read())
    elif content:
        data = b64encode(content)

    return {
        "content": str(data, 'utf-8'),
        "encoding": 'base64',
        "extension": ext,
        "mimetype": mime_type,
        "size": size
    }


if __name__ == '__main__':
    import cas.configure as config

    config.configure_logging(['pika'])
    try:
        raise KeyError('I know')
    except KeyError as e:
        import traceback

        st = traceback.format_exc()
        print('done')



import logging
import threading

_LOGGER = logging.getLogger('cas.model.concurrency')


class ResultThread(threading.Thread):

    @property
    def result(self):
        return self.__result

    def __init__(self, target, *args, **kwargs):
        if not callable(target):
            raise TypeError('expected callable function, not {0}'.format(type(target)))

        super(ResultThread, self).__init__(target=self.execute)
        self.__args = args
        self.__kwargs = kwargs
        self.__exec_method = target
        self.__result = None

    def execute(self):
        self.__result = self.__exec_method(*self.__args, **self.__kwargs)


class RetryThread(ResultThread):

    def __init__(self, target, retry_callback=None, retry_count=5, *args, **kwargs):
        if not callable(target):
            raise TypeError('expected callable function, not {0}'.format(type(target)))

        if retry_callback is not None and not callable(retry_callback):
            raise TypeError('expected callable function, not {0}'.format(type(retry_callback)))

        super(RetryThread, self).__init__(target, *args, **kwargs)
        self.__retry_callback = retry_callback
        self.__current_attempt = 0
        self.__max_retry = retry_count + 1

    def execute(self):
        while self.__current_attempt < self.__max_retry:
            try:
                _LOGGER.debug(f'Retry attempt {self.__current_attempt}')
                super(RetryThread, self).execute()
                if hasattr(self, '_RetryThread__current_exception'):
                    del self.__current_exception

                break
            except Exception as e:
                _LOGGER.exception(f'Retry Thread {threading.current_thread().name} Failed - '
                                  f'Attempt {self.__current_attempt}', exc_info=e)

                # noinspection PyAttributeOutsideInit
                self.__current_exception = e
                if self.__retry_callback is not None:
                    self.__retry_callback(self.__current_attempt, e)

                self.__current_attempt += 1
                continue

        if hasattr(self, '_RetryThread__current_exception'):
            raise self.__current_exception


class CallBackThread(RetryThread):

    def __init__(self, target, *args, callback=None, exception_callback=None, retry_callback=None, retry_count=5,
                 **kwargs):
        if callback is not None and not callable(callback):
            raise TypeError('expected callable function, not {0}'.format(type(callback)))

        if exception_callback is not None and not callable(exception_callback):
            raise TypeError('expected callable function, not {0}'.format(type(exception_callback)))

        super(CallBackThread, self).__init__(target, retry_callback, retry_count,
                                             *args, **kwargs)
        self.__callback = callback
        self.__exception_callback = exception_callback

    def execute(self):
        try:
            super(CallBackThread, self).execute()
            if self.__callback is not None:
                self.__callback(self.result)
        except Exception as e:
            if self.__exception_callback is not None:
                self.__exception_callback(e)
            else:
                raise e


# class CallBackThread(threading.Thread):
#     """A thread that contains a callback when the thread has finished running"""
#
#     def __init__(self, target, *args, callback=None, exception_callback=None, retry_callback=None, retry_count=5,
#                  **kwargs):
#         if not callable(target):
#             raise TypeError('expected callable function, not {0}'.format(type(target)))
#
#         if callback is not None and not callable(callback):
#             raise TypeError('expected callable function, not {0}'.format(type(callback)))
#
#         if exception_callback is not None and not callable(exception_callback):
#             raise TypeError('expected callable function, not {0}'.format(type(exception_callback)))
#
#         if retry_callback is not None and not callable(retry_callback):
#             raise TypeError('expected callable function, not {0}'.format(type(retry_callback)))
#
#         super(CallBackThread, self).__init__(target=self.__callback_target)
#         self.__args = args
#         self.__kwargs = kwargs
#         self.__callback = callback
#         self.__exception_callback = exception_callback
#         self.__retry_callback = retry_callback
#         self.__exec_method = target
#         self.__current_attempt = 0
#         self.__max_retry = retry_count + 1
#
#     def __callback_target(self):
#
#         while self.__current_attempt < self.__max_retry:
#             try:
#                 _LOGGER.debug(f'Retry attempt {self.__current_attempt}')
#                 self.__exec_method(*self.__args, **self.__kwargs)
#                 if hasattr(self, '_CallBackThread__current_exception'):
#                     del self.__current_exception
#                 if self.__callback is not None:
#                     self.__callback()
#
#                 break
#             except Exception as e:
#                 _LOGGER.exception(f'Callback Thread {threading.current_thread().name} Failed - '
#                                   f'Attempt {self.__current_attempt}', exc_info=e)
#
#                 self.__current_exception = e
#                 if self.__retry_callback is not None:
#                     self.__retry_callback(self.__current_attempt, e)
#
#                 self.__current_attempt += 1
#                 continue
#
#         if hasattr(self, '_CallBackThread__current_exception'):
#             if self.__exception_callback is not None:
#                 self.__exception_callback(self.__current_exception)
#             else:
#                 raise self.__current_exception

if __name__ == '__main__':
    a = -1


    def call():
        global a
        a = a + 1
        if a == 0:
            raise ValueError('')
        return 'Hello'


    c = CallBackThread(call)
    c.start()
    c.join()
    print(c.result)



import datetime


def current_time_str():
    return datetime.datetime.utcnow().strftime("%Y-%m-%dT%H:%M:%SZ")


def get_variable(container, var_name):
    var_names = var_name if isinstance(var_name, list) and all(isinstance(e, str) for e in var_name) \
        else var_name.split('.')
    current = None
    if isinstance(container, dict) and var_names[0] in container:
        current = container[var_names.pop(0)]
    elif hasattr(container, var_names[0]):
        current = getattr(container, var_names.pop(0))

    if current is not None:
        if len(var_names) == 0:
            return current
        else:
            return get_variable(current, var_names)
    else:
        return 'NotFound'

def has_value(val: str) -> bool:
    return val is not None and len(val) > 0



# noinspection PyUnresolvedReferences
import encodings.idna
import logging
import os
import shutil
import uuid
from enum import IntEnum, unique
from pathlib import Path

import boto3
from botocore.exceptions import ClientError

from cas.concurrency import CallBackThread

_LOGGER = logging.getLogger('cas.model.file')
_BUCKET_NAME = os.getenv('AWS_CURRENT_BUCKET_NAME', 'acabim-testing')


@unique
class AwsFileType(IntEnum):
    CAP = 1  # do not change int values as they are in sync with the .NET clients
    LKM = 2
    LRML = 3
    IFC = 4
    RVT = 5
    BCF = 6
    CSV = 7
    PNG = 8
    PDF = 9
    JPEG = 10
    SVG = 11
    GLB = 12
    JSON = 13
    ZIP = 14,
    HTML = 15,
    Unknown = 0

    @property
    def default_file_format(self):
        return self.name.lower()


@unique
class AwsFileCategory(IntEnum):
    General = 0
    BuildingData = 1
    CapReport = 2
    FacilityManagement = 3
    OrganisationData = 4
    Log = 10


class AwsFileDownloader:
    """This class is used to retrieve a file from the AWS S3. When this class goes out of scope, and is garbage
    collected, the downloaded file will automatically be removed from the temporary directory"""

    def __init__(self, file_name, file_type, file_category, auto_download=True, file_available_callback=None,
                 file_unavailable_callback=None):
        if file_available_callback is not None and not callable(file_available_callback):
            raise TypeError('expected callable function, not {0}'.format(type(file_available_callback)))

        if file_unavailable_callback is not None and not callable(file_unavailable_callback):
            raise TypeError('expected callable function, not {0}'.format(type(file_unavailable_callback)))

        self.file_name = file_name
        self.file_type = file_type
        self.file_category = file_category
        self.__download_thread = None
        self.__temp_folder = str(uuid.uuid4())
        self.__file_available_callback = file_available_callback
        self.__file_unavailable_callback = file_unavailable_callback

        _LOGGER.debug('Initialised S3 Access for AWS File "%s"', file_name)
        _LOGGER.debug('Will be downloaded to path "%s"', self.get_local_file_path())
        if auto_download:
            self.download_file()

    def __del__(self):
        self.__cleanup()

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.__cleanup()

    def is_file_downloaded(self):
        return os.path.exists(self.get_local_file_path())

    def is_downloading(self):
        return self.__download_thread is None

    def get_local_file_path(self):
        """Returns the path of the file which will be / was downloaded to"""
        return os.path.join(self.get_local_file_dir(), self.get_local_file_name())

    def get_local_file_name(self):
        """Returns the file name which will be / was downloaded to"""
        f_format = self.file_type.default_file_format
        return f'{self.file_name}.{f_format}' if not self.file_name.endswith(f'.{f_format}') else self.file_name

    def get_local_file_dir(self):
        """Returns the directory to which the file will be downloaded to"""
        return self.__get_temp_path()

    def wait_for_download(self):
        """Wait for the download to complete"""
        if self.__download_thread is not None:
            self.__download_thread.join()

    def download_file(self):
        if not self.is_file_downloaded():
            if self.__download_thread is None:
                path = generate_bucket_path(self.file_category, self.file_type)

                # if the file_name doesn't contain the file format then we add it here, otherwise we don't
                f_format = self.file_type.default_file_format
                f_path = os.path.join(path, self.file_name)
                bucket_path = f'{f_path}.{f_format}' if not self.file_name.endswith(f'.{f_format}') else f_path

                file_path = self.get_local_file_path()

                _LOGGER.info(f'Using S3 Bucket {_BUCKET_NAME} for downloading {bucket_path}')
                self.__download_thread = CallBackThread(AwsFileDownloader.__download_impl, _BUCKET_NAME, bucket_path,
                                                        file_path, callback=self.__on_download_finished,
                                                        exception_callback=self.__on_download_exception,
                                                        retry_callback=self.__on_download_retry, retry_count=5)
                self.__download_thread.setName('cas_aws_file_dl:{0}'.format(self.file_name))
                self.__download_thread.daemon = True
                self.__download_thread.start()
                _LOGGER.debug('File download started for "%s"', self.file_name)
            else:
                _LOGGER.warning('File is already downloading')
        else:
            _LOGGER.debug('File is already downloaded')

    def __get_temp_path(self):
        folder_path = os.path.join(str(Path.home()), '.local', 'acabim-python', self.__temp_folder)
        if not os.path.exists(folder_path):
            _LOGGER.debug('Creating Path "%s"', folder_path)
            os.makedirs(folder_path)

        return folder_path

    def __on_download_finished(self, _):
        _LOGGER.info('File "%s" download complete', self.file_name)
        self.__download_thread = None
        if self.__file_available_callback is not None:
            self.__file_available_callback(self)

    def __on_download_exception(self, exception):
        self.__download_thread = None
        if self.__file_unavailable_callback is not None:
            self.__file_unavailable_callback(exception)
        else:
            _LOGGER.error(f'Unable to download file {self.file_name}')
            raise exception

    def __on_download_retry(self, attempt, exception):
        _LOGGER.warning(f'File download {attempt} failed: {str(exception)}')
        self.__cleanup()  # remove any fragments of the file which may have been downloaded before
        self.__get_temp_path()  # so that we recreate the folder path if it was removed during cleanup

    def __cleanup(self):
        _LOGGER.debug('Removing downloaded file "%s" (if available)', self.file_name)
        try:
            if self.is_file_downloaded():
                os.remove(self.get_local_file_path())

            t_path = self.__get_temp_path()
            _LOGGER.debug(f'Removing Downloader Temporary Path: {t_path}')
            shutil.rmtree(t_path, ignore_errors=True)
        except OSError as e:
            _LOGGER.exception('Unable to cleanup data', exc_info=e)

    @staticmethod
    def __download_impl(bucket_name, bk_path, f_path):
        _LOGGER.debug(f'Starting S3 File download from bucket {bucket_name} : Path: {bk_path}')
        _LOGGER.debug(f'Downloading to cache file path {f_path}')
        if not os.path.exists(Path(f_path).parent.absolute()):
            _LOGGER.debug('Creating Path %s', Path(f_path).parent.absolute())
            os.makedirs(Path(f_path).parent.absolute(), exist_ok=True)
        boto3.client('s3').download_file(Bucket=_BUCKET_NAME, Key=bk_path, Filename=f_path)


def upload_file(file_path, file_type: AwsFileType, file_category: AwsFileCategory):
    file_name = os.path.basename(file_path)
    obj_name = '{0}/{1}'.format(generate_bucket_path(file_category, file_type), file_name)
    try:
        boto3.client('s3').upload_file(file_path, _BUCKET_NAME, obj_name)
        _LOGGER.info('File %s uploaded to %s successfully', file_name, _BUCKET_NAME)
    except ClientError as e:
        _LOGGER.exception(e, 'Unable to upload file %s to bucket %s', file_path, _BUCKET_NAME)
        return False
    return True


def delete_file(file_name, file_type: AwsFileType, file_category: AwsFileCategory):
    """Deletes the given file from the AWS S3 bucket. The file path will automatically be created from the file_type
    and the file_category"""
    # if the file_name doesn't contain the file format then we add it here, otherwise we don't
    f_format = file_type.default_file_format
    normalised_name = f'{file_name}.{f_format}' if not file_name.endswith(f'.{f_format}') else file_name
    obj_name = '{0}/{1}'.format(generate_bucket_path(file_category, file_type), normalised_name)

    try:
        boto3.client('s3').delete_object(Bucket=_BUCKET_NAME, Key=obj_name)
        _LOGGER.info('File %s deleted from %s successfully', file_name, _BUCKET_NAME)
    except ClientError as e:
        _LOGGER.exception(e, 'Unable to delete file %s from bucket %s', obj_name, _BUCKET_NAME)
        return False
    return True


def exists_file(file_name, file_type: AwsFileType, file_category: AwsFileCategory):
    """Check if a File with the specified name, file type and file category exists in the S3 bucket"""
    obj_name = '{0}/{1}.{2}'.format(generate_bucket_path(file_category, file_type), file_name,
                                    file_type.default_file_format)
    try:
        res = boto3.client('s3').list_objects_v2(Bucket=_BUCKET_NAME, Prefix=obj_name, MaxKeys=1)
        for obj in res.get('Contents', []):
            if obj['Key'] == obj_name:
                return True
    except ClientError as e:
        _LOGGER.exception('Unable to check file availability: Error', exc_info=e)

    return False


def generate_bucket_path(file_category, file_type):
    if file_category is AwsFileCategory.General:
        return file_type.name
    else:
        return '{0}/{1}'.format(file_category.name, file_type.name)


if __name__ == '__main__':
    import cas.configure as c

    c.configure_logging()
    # d = AwsFileDownloader('Linwood', AwsFileType.IFC, AwsFileCategory.BuildingData)
    # d.wait_for_download()
    f = exists_file('Linwood', AwsFileType.IFC, AwsFileCategory.BuildingData)
    _LOGGER.debug(f)
    pass



import os
from sqlalchemy.orm import sessionmaker
from sqlalchemy import create_engine

db_username = os.getenv('DATABASE_USERNAME', 'SA')
db_password = os.getenv('DATABASE_PASSWORD', 'cas-dev.Local')
db_host = os.getenv('DATABASE_HOST', 'localhost')
db_name = os.getenv('DATABASE_NAME', 'acabim_common')

db_url = f'mssql+pyodbc://{db_username}:{db_password}@{db_host}/{db_name}?driver=ODBC+Driver+17+for+SQL+Server'

sql_engine = create_engine(db_url, echo=True)
Session = sessionmaker(bind=sql_engine)
